#!/usr/bin/env python
# encoding: utf8
'''
Usage: verify [options] <filename>

Options:
    --validation=<rules>
        validation definition file in YAML format
        [default: verify.yaml]

    -h, --help      print this help text
    -v, --verbose   give more details about the processing
'''

from docopt import docopt
import logging
from ruamel import yaml
import re

log = logging.getLogger()
logging.basicConfig(level=100) #logging.INFO)

def check(condition, warning_if_false, level=logging.WARNING):
    log.debug("check for '%s'", warning_if_false)
    if not condition:
        if level > logging.WARNING:
            raise AssertionError(warning_if_false)
        log.log(level, warning_if_false)

class Validator(object):
    def __init__(self, rulefile=None, rules=None, throw=False):
        '''
        Initialize the validator with given `rules` or rules loaded from given
        YAML `rulefile`.

        If parameter `throw` is `True`, then validate() will raise an
        AssertionError with additional information instead of True or False.
        '''
        self.throw = throw
        self.rules = rules
        if rulefile is not None:
            log.debug("opening %s...", rulefile)
            with open(rulefile, 'r') as infile:
                self.rules = yaml.safe_load(infile)

    def validate_file(self, filename):
        '''
        Try to load yaml from the file with given name and run its data through validate

        returns the result of validate()
        '''
        log.debug("opening %s...", filename)
        data = None
        with open(filename, 'r') as infile:
            data = yaml.safe_load(infile)
        return self.validate(data)

    def _validate_dict(self, item, rules, path):
        '''
        Validate a dict item. Optional key `regex` can be used to verify the
        content of the string

        >>> v = Validator(rules={'type': 'dict',
        ...                      "keys": {'type': 'str', 'regex':r'\w+'},
        ...                      "values": {'type': 'int', 'min': 0},
        ...                     }
        ...              )
        >>> v.validate({"a", 42})
        False
        >>> v.validate({"a": 42})
        True
        >>> v.validate({"a 4": 42})
        False
        >>> v.validate({"b": "42"})
        False
        >>> v.validate({"4": -42})
        False
        '''
        assert isinstance(item, dict), "%s: expected dict, got %s!" % (":".join(path), type(item).__name__)
        checktype = "dict"
        if "keys" in rules:
            log.debug("assert correct type of keys of %s...", ":".join(path))
            for k in item.keys():
                assert self.validate(k, rules['keys'], path+[str(k)])
        if "values" in rules:
            log.debug("assert correct values of %s: %s", ":".join(path), checktype)
            for v in item.values():
                assert self.validate(v, rules['values'], path+["values", str(v)])
        return True

    def _validate_str(self, item, rules, path):
        '''
        Validate a string item. Optional key `regex` can be used to verify the
        content of the string

        >>> v = Validator(rules={'type': 'str'})
        >>> v.validate("1")
        True
        >>> v.validate(1)
        False
        >>> v = Validator(rules={'type': 'str', 'regex':r'\d+\s?units'})
        >>> v.validate("23 units")
        True
        >>> v.validate("unit 23")
        False
        '''
        assert isinstance(item, str), "%s: expected str, got %s!" % (":".join(path), type(item).__name__)
        if "regex" in rules:
            assert re.fullmatch(rules['regex'], item, rules.get('regex_flags', 0)), "%s INVALID" % ":".join(path)
        return True

    def _validate_int(self, item, rules, path):
        '''
        Validate an integer.

        Optional keys 'min' and 'max' can be used to limit allowed values. The
        range is inclusive `[min, max]`. The optional keys 'sup', 'inf' can be
        used for the exclusive range checks `(inf, sup)`. Range can be mixed
        `[min, sup)`. Range checks are independant, so empty ranges (min=1,
        max=0) are possible.

        >>> v = Validator(rules={'type': 'int'})
        >>> v.validate(1)
        True
        >>> v.validate(3.1415)
        False
        >>> v = Validator(rules={'type': 'int', 'min':42, 'sup':43})
        >>> v.validate(41)
        False
        >>> v.validate(42)
        True
        >>> v.validate(43)
        False
        >>> v = Validator(rules={'type': 'int', 'inf':41, 'max':42})
        >>> v.validate(41)
        False
        >>> v.validate(42)
        True
        >>> v.validate(43)
        False
        '''
        assert isinstance(item, int), "%s: expected int, got %s!" % (":".join(path), type(item).__name__)
        if "min" in rules:
            assert item >= rules['min']
        if "inf" in rules:
            assert item > rules['inf']
        if "sup" in rules:
            assert item < rules['sup']
        if "max" in rules:
            assert item <= rules['max']
        return True

    def _validate_float(self, item, rules, path):
        '''
        Validate a floating point number.

        Optional keys 'min' and 'max' can be used to limit allowed values. The
        range is inclusive `[min, max]`. The optional keys 'sup', 'inf' can be
        used for the exclusive range checks `(inf, sup)`. Range can be mixed
        `[min, sup)`. Range checks are independant, so empty ranges (min=1,
        max=0) are possible.

        >>> v = Validator(rules={'type': 'float'})
        >>> v.validate(1)
        False
        >>> v.validate(3.1415)
        True
        >>> v = Validator(rules={'type': 'float', 'min':42, 'sup':43})
        >>> v.validate(41.)
        False
        >>> v.validate(42.)
        True
        >>> v.validate(43.)
        False
        >>> v = Validator(rules={'type': 'float', 'inf':41, 'max':42})
        >>> v.validate(41.)
        False
        >>> v.validate(42.)
        True
        >>> v.validate(43.)
        False
        '''
        assert isinstance(item, float), "%s: expected float got %s!" % (":".join(path), type(item).__name__)
        if "min" in rules:
            assert item >= rules['min']
        if "inf" in rules:
            assert item > rules['inf']
        if "sup" in rules:
            assert item < rules['sup']
        if "max" in rules:
            assert item <= rules['max']
        return True

    def _validate_in(self, item, rules, path):
        '''
        Branch validation to satisfy any of a given list of rules

        >>> v = Validator(rules={'type': 'in',
        ...                      'values': [{'type': 'int', 'min': 23},
        ...                                 {'type': 'float', 'max': 1.0},
        ...              ]})
        >>> v.validate(1)
        False
        >>> v.validate(1.0)
        True
        >>> v.validate(42)
        True
        >>> v.validate(42.0)
        False
        '''
        ###
        ### NOTE: Since this is an idirect validation calling self.validate()
        ### the result must be asserted in order to be independant of self.throw.
        ###
        checkers = rules['values']
        assert isinstance(checkers, list), "checker of type 'in' must have 'values'"
        for rule in checkers:
            try:
                assert self.validate(item, rule, path)
                return True
            except AssertionError:
                pass
        assert False, "%s: got %s, expected %s" % (":".join(path), type(item).__name__, str(rules))

    def _validate_list(self, item, rules, path):
        '''
        >>> v = Validator(rules={'type': 'list'})
        >>> v.validate([1,2,3])
        True
        >>> v.validate("str")
        False
        >>> v = Validator(rules={'type': 'list', "values": {'type': "int"}})
        >>> v.validate([1,2,3])
        True
        >>> v.validate([1,'str',3])
        False
        '''
        assert isinstance(item, list), "%s: expected list, got %s!" % (":".join(path), type(item).__name__)
        if "values" in rules:
            for i, subitem in enumerate(item):
                assert self.validate(subitem, rules['values'], path+[str(i)])
        return True

    def _validate_dictdescent(self, item, rules, path):
        '''
        >>> v = Validator(rules={'type': 'dictdescent'})
        >>> v.validate([1,2,3])
        False
        >>> v = Validator(rules={'type': 'list', "values": {'type': "int"}})
        >>> v.validate([1,2,3])
        True
        >>> v.validate([1,'str',3])
        False
        '''
        assert isinstance(item, dict), "%s: expected dict, got %s!" % (":".join(path), type(item).__name__)
        if "mandatory" in rules:
            assert all([key in item for key in rules['mandatory']]), "%s: misses mandatory keys" % (":".join(path))
        #if "allowed" in rules:
        #   good
        if "deprecated" in rules:
            for key in rules['deprecated'].keys():
                if key in item:
                    log.warning("%s: deprecated key %s", ":".join(path), key)
        if "forbidden" in rules:
            assert not any([key in item for key in rules['forbidden'].keys()]), "%s: has forbidden keys" % (":".join(path))

        otherkeys = [key for key in item.keys() if not any([key in rules.get(good, {}) for good in ['mandatory', 'allowed', 'deprecated']])]
        assert rules.get("others-allowed", False) or not any(otherkeys), "%s: there are keys that are addtional keys %s and others-allowed is False" % (":".join(path), repr(otherkeys))

        return True

    def validate(self, tree, rules=None, path=['/']):
        '''
        Validate the object `tree` according to given rules. By default, if no
        rules are given, the rule base loaded during the construction of the
        validator is used.

        >>> v = Validator(rules={'type': 'list'}, throw=False)
        >>> v.validate(123)
        False
        >>> v = Validator(rules={'type': 'list'}, throw=True)
        >>> v.validate(123)
        Traceback (most recent call last):
            ...
        AssertionError: /: expected list, got int!
        '''
        if rules is None: rules = self.rules
        if not isinstance(rules, dict):
            print("rules = %s", repr(rules))
            assert False, "rule must be of type dict, but is %s" % repr(rules)
        checktype = rules.get('type')
        try:
            check = getattr(self, "_validate_"+checktype)
        except Exception:
            log.error("no validator of type '%s'", checktype)
            raise
        try:
            log.debug("assert correct type %s: %s", ":".join(path), checktype)
            assert check(tree, rules, path), "validation of %s failed!" % checktype
        except AssertionError as e:
            log.error("%s IS INVALID: %s", ":".join(path), e)
            if self.throw:
                raise
            return False
        log.debug("%s validated as %s", ":".join(path), checktype)
        return True

if __name__ == '__main__':
    args = docopt(__doc__)
    if args['--verbose']:
        log.setLevel(logging.DEBUG)

    validator = Validator(rulefile = args['--validation'])

    valid = validator.validate_file(args['<filename>'])
    if valid:
        print("valid.")
    else:
        print("INVALID!")




#!/usr/bin/env python
# encoding: utf8
'''
Usage: reporter [options] [-y|-p|-j] [--list=<name>...] [--filter=<pyfile>...] [--template=<template>] [--output=<file>] [<datadef>...]

   Create a report by filling the template with data according to <datadef>.

   Each <datadef> must be a pair "<key>=<filename>", where key is the
   identifier used in the template and <filename> is the YAML file containing
   the corresponding data.

Options:
    -t, --template=<template>
        use an alternative template [default: report.md]

    --template-dir=<dir>
        specify the base directory for the used templates [default: .]

    --list=<name>...
        all names given in --list options will be initialized as empty list and
        consecutive `datadef`s will be appended to the list instead of assigned
        to the name. Example:

           reporter --list=bar foo=a.yml bar=b.yml bar=c.yml

        would give objects `foo` and `bar`, where `foo` represents the object
        loaded from `a.yml` and `bar` is a list containting objects loaded from
        `b.yml` and `c.yml`.

        This option can be given more than once to create multiple lists, e.g.

           reporter --list=bar --list=baz  bar=a.yml bar=b.yml baz=c.yml baz=d.yml

        would give lists `bar = [a, b]` and `baz = [c, d]`.

    --meta-dict=<name>
        change name of the metadata dictionary [default: report]

    -y, --yaml
        use yaml.load (default)

    --yaml-loader=<type>
        choose a specific loader type from the ruamel lib [default: safe]

    -j, --json
        use json.load instead of yaml.load

    -p, --python
        use eval() loading (very insecure!!!)

    -o, --output=<filename>
        write to this file instead of stdout

    --filter=<pythonfile>
        load functions from given python file and add them to the available
        filters

    -v, --verbose       increase output
    -h, --help          print this text
'''
from docopt import docopt

from pprint import pformat
import logging
from ruamel.yaml import YAML
from jinja2 import Environment, FileSystemLoader, select_autoescape
import os.path
import shlex
import sys
import datetime
log = logging.getLogger()
logging.basicConfig(level=logging.INFO)

def pythoneval(stream):
    return eval(stream.read())

def loadfilters(listoffiles):
    import importlib.util
    filters = dict()
    for filename in listoffiles:
        log.debug("loading filters from %s...", filename)
        spec = importlib.util.spec_from_file_location("loaded_filters", filename)
        assert spec, "Could not load python file %s" % filename
        newfilters = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(newfilters)
        for x in dir(newfilters):
            if x.startswith('_'): continue
            assert x not in filters
            filters[x] = getattr(newfilters, x)
            log.debug("   filter %s", x)
    return filters

def main(cmdline = None):
    if cmdline is None:
        cmdline = sys.argv[1:]
    else:
        cmdline = shlex.split(cmdline)  # have a possibility for testing
    args = docopt(__doc__, cmdline)

    if args['--verbose']:
        log.setLevel(logging.DEBUG)
    log.debug(pformat(args))

    log.debug("template directory '%s'", args['--template-dir'])
    env = Environment(
        loader=FileSystemLoader(args['--template-dir']),
        trim_blocks=True,
        lstrip_blocks=True,
        #autoescape=select_autoescape(['html', 'xml'])
        extensions=[
            'jinja2.ext.loopcontrols',
        ],
    )
    newfilters = loadfilters(args['--filter'])
    log.debug("new filters: %s", newfilters)
    env.filters.update(newfilters)

    # load template
    log.info("loading template '%s'...", args['--template'])
    tmpl = env.get_template(args['--template'])

    yamlin = YAML(typ=args['--yaml-loader'])
    dataloader = yamlin.load
    if args['--yaml']: dataloader = yamlin.load
    if args['--python']: dataloader = pythoneval
    if args['--json']:
        import json
        dataloader = json.load

    # load data
    data = dict()
    for datadef in args['<datadef>']:
        dataid, datafilename = datadef.split("=", 1)
        if dataid in args['--list']:
            log.info("loading '%s' entry from '%s'...", dataid, datafilename)
            data.setdefault(dataid, list()).append(dataloader(open(datafilename, 'r')))
        else:
            log.info("loading '%s' from '%s'...", dataid, datafilename)
            data[dataid] = dataloader(open(datafilename, 'r'))
    log.debug("loading data complete.")

    if args['--meta-dict'] in data:
        log.error("metadata dictionary has same name as loaded data! use '--meta-dict' to rename, or change datadef")
        return 1

    data[args['--meta-dict']] = {
        "time": datetime.datetime.now(),
    }

    # output result
    ostream = sys.stdout
    if args['--output'] is not None:
        ostream = open(args['--output'], 'w')
    log.info("writing output to %s", ostream.name)
    with ostream as outfile:
        outfile.write(tmpl.render(**data))

    return 0

if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-

from setuptools import setup
import os
import sys
try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

long_description = open("README.md").read()
install_requires = [
    'pandoc',
    'jinja2',
    'docopt',
    'ruamel-yaml',
]

setup(
    name="datareport",
    version='0.1.7',
    packages=['datareport'],
    package_data={
        '': [
            'LICENSE.txt'
        ],
    },

    install_requires=install_requires,

    author="Dennis Terhorst",
    author_email="d.terhorst@fz-juelich.de",
    description="Tool to combine YAML datasets with Jinja2-based templates.",
    long_description=long_description,

    entry_points = {
        'console_scripts': [
            'reporter = datareport.reporter:main',
        ],
    },

    # https://opensource.org/licenses/BSD-2-Clause
    license="BSD",

    url='https://github.com/INM-6/datareport',
    # https://pypi.org/pypi?:action=list_classifiers
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering']
)
